#
#    forest/dbm.py implements a dbm backend
#    Copyright (C) 2002  Michael Urman
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
dbm.py

Implements a dbm file backend in the forest messaging system
"""

import backend
import forest   # load configuration

import os
import anydbm
import time

def file_to_name(filename):
    """file_to_name(filename) -> name
    
    Return the name associated with the filename by turning directory
    separators into name separators, and removing the base directory if
    possible"""

    if filename.startswith(forest.root):
        name = filename[len(forest.root):]
    name = forest.spliton.join(name.split(os.sep))
    return name

def name_to_file(name):
    """name_to_file(name) -> filename

    Return the filename associated to a tree name by turning the separation
    character into directory separators, and prepending the base directory
    if necessary"""

    if os.path.splitdrive(name)[0] or name.startswith(os.sep):
        return os.path.normpath(tree)
    elif forest.spliton:
        dirs = [forest.root]
        dirs.extend(tree.split(forest.spliton))
        file = os.path.normpath(os.path.join(*dirs))
    else:
        file = os.path.normpath(os.path.join(forest.root, tree))
    # print 'tree:', tree, '->', file + forest.ext
    return file + forest.ext

class error(backend.error):
    pass


class trees:
    """Represent the trees of the forest with auto-opening"""

    def __init__(self):
        """__init__(self) -> trees

        Build a list of available trees"""

        self.trees = {}
        os.path.walk(forest.root, self.add_to_list, None)

    def has_key(self, key):
        """has_key(self, key) -> bool

        Return if tree key is in the dictionary"""

        return self.trees.has_key(key)

    def add_to_list(self, arg, directory, files):
        """add_to_list(self, arg, directory, files) -> None
        
        Help build a list of trees"""

        for file in files:
            if not os.path.isdir(file):
                self.trees[name] = tree(self.file_to_name(file), cache=0)

    def __contains__(self, name):
        return self.tree.has_key(name)

    def __getitem__(self, name):
        """trees[name] -> tree

        Return or open and return a tree by name"""

        try:
            return self.tree[name]
        except KeyError:
            me = self.trees[name] = tree(name)
            return me


class tree:
    """Represent a tree with corresponding convenience functions"""

    def __init__(self, name, cache=1):
        self.name = name
        self.file = name_to_file(name)
        self.access = time.time()
        try:
            self.db = anydbm.open(self.file, 'w')
        except anydbm.error:
            raise error("tree '%s' (%s) could not be opened" % (name, self.file))
        self.modified = db['modified']
        self.branches = branches(self)

        if not cache:
            self.db.close()
            self.db = None

    def close_if_stale(self, now):
        """close_if_stale(self, now) -> None

        Close the dbm handle if tree hasn't been accessed in over
        forrest.close_dbm_timeout seconds"""

        if self.access + forest.close_dbm_timeout < now:
            self.db.close()
            self.db = None

    def set_blurb(self, blurb):
        self.db['blurb'] = blurb

    def set_attention_blurb(self, blurb):
        self.db['attention'] = blurb


class branches:
    """Represent the branches of the tree with auto-opening"""

    def __init__(self, tree):
        self.tree = tree
        self.branches = {}
        self.branch_count = tree.db['branch_count']

    def __len__(self):
        return self.branch_count

    def __getitem__(self, index):
        """branches[index] -> branch

        Return or open and return a branch by index"""
        try:
            return self.branches[index]
        except KeyError:
            if 0 <= index < self.branch_count:
                me = self.branches[index] = branch(self.tree, index)
                return me
            else:
                raise IndexError("Branch index %s out of range" % index)

    def post(self, post):
        """post(self, post) -> int

        Post a new branch with post; return branch index"""

        raise NotImplementedError("hahaha")


class branch:
    """Represent a branch with corresponding convenience functions"""

    def __init__(self, tree, index):
        self.tree = tree
        self.index = index
        self.post_count = db['branch %d post count' % index]
        self.posts = {}

    def __len__(self):
        return self.post_count

    def __getitem__(self, index):
        """branches[index] -> branch

        Return or open and return a post by index"""

        try:
            return self.posts[index]
        except KeyError:
            if 0 <= index < self.post_count:
                me = self.posts[index] = post(self.tree, self.index, index)
                return me
            else:
                raise IndexError("Post index %s out of range on branch %s" %
                    (index, self.index)

    def post(self, post):
        """post(self, post) -> int

        Post a new branch with post; return branch index"""

        raise NotImplementedError("hahaha")


class post:
    """Represent a post: text and meta information"""

    def __init__(self, name=None, uid=None,
                 blurb=None, text=None, posted_at=None):
        self.name = name
        self.uid = uid
        self.blurb = blurb
        self.text = text
        self.time = posted_at or time.time()

class dbm(backend.backend):

    def __init__(self):
        if not os.path.exists(forest.root):
            print "Fatal Error: root path '%s' does not exist." % forest.root
        elif not os.path.isdir(forest.root):
            print "Fatal Error: root path '%s' is not a directory." % forest.root
        elif not os.access(forest.root, os.R_OK | os.W_OK | os.X_OK):
            print "Fatal Error: missing privileges on root path '%s'." % forest.root
        self.trees = trees()

    # get info
    def get_trees(self): pass
    def count_tree_branches(self, tree): pass
    def get_tree_branches(self, tree): pass
    def count_tree_branch_posts(self, tree, branch): pass
    def get_tree_branch_posts(self, tree, branch): pass
    def read_tree_branch_post(self, tree, branch, post): pass

    # set info
    def set_tree_blurb(self, tree, blurb): pass
    def set_tree_attention_blurb(self, tree, blurb): pass

    # write info
    def make_tree(self, name):
        """make_tree(self, name) -> None

        Create a new tree and cache it"""

        if self.trees.has_key(name):
            raise error("tree '%s' has already been opened." % name)

        filename = self.name_to_file(name)
        dirname = os.path.dirname(filename)

        # create parent path if necessary
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        # create
        try:
            db = anydbm.open(file, 'c')
        except anydbm.error:
            raise error("tree '%s' (%s) could not be created." %
                (name, filename))
        if db.has_key('created'):
            raise error("tree '%s' (%s) already exists." %
                (name, filename))
        when = str(time.time())
        db['created'] = when
        db['modified'] = when
        db['branches'] = str(0)

        db.close()
        tree = self.trees[name]
        tree.set_blurb('About ' + name)
        tree.set_attention_blurb('** DIRECTOR **')

    def make_post(self, tree, branch, post):
        self.trees[tree].branches[branch].post(post)

    def make_branch_post(self, tree, post, blurb):
        self.trees[tree].branches.post(post, blurb)

    # hide info
    def hide_tree(self, tree): pass
    def hide_tree_branch(self, tree, branch): pass
    def hide_tree_branch_post(self, tree, branch, post): pass

    # delete info
    def delete_tree(self, tree): pass
    def delete_tree_branch(self, tree, branch): pass
    def delete_tree_branch_post(self, tree, branch, post): pass

    # privileges
    def set_global_priv(self, priv): pass
    def set_tree_priv(self, tree, priv): pass
    def read_global_priv(self, user): pass
    def read_tree_priv(self, tree, user): pass
